import six.moves
from numbers import Integral
from .iter_dispatch import range_iterator

__all__ = ['xrange']


def _check_integral(value):
    if not isinstance(value, Integral):
        raise TypeError("'{}' object cannot be interpreted "
                        "as an integer".format(type(value).__name__))


class xrange(object):
    """A replacement for Python 3 `range()` (and Python 2 `xrange()`) that
    yields picklable iterators when iterated upon.
    """
    __slots__ = ['_start', '_stop', '_step']

    def __init__(self, *args):
        self._start = 0
        self._step = 1
        if len(args) == 0:
            raise TypeError("{} expected 1 arguments, got 0".format(
                self.__class__.__name__))
        elif len(args) == 1:
            self._stop = args[0]
            self._start = 0
        elif len(args) >= 2:
            self._start = args[0]
            self._stop = args[1]
        if len(args) == 3:
            self._step = args[2]
        if len(args) > 3:
            raise TypeError("{} expected at most 3 arguments, got {}".format(
                self.__class__.__name__, len(args)))
        _check_integral(self._start)
        _check_integral(self._stop)
        _check_integral(self._step)

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def step(self):
        return self._step

    def count(self, i):
        """rangeobject.count(value) -> integer -- return number of occurrences
        of value
        """
        if self._stop > self._start and self._step > 0:
            return int(self._start <= i < self._stop and
                       (i - self._start) % self._step == 0)
        elif self._stop < self._start and self._step < 0:
            return int(self._start >= i > self._stop and
                       (i - self._start) % self._step == 0)
        else:
            return False

    def index(self, i):
        """xrangeobject.index(value, [start, [stop]]) -> integer --
        return index of value.
        Raise ValueError if the value is not present.
        """
        if self.count(i) == 0:
            raise ValueError("{} is not in range".format(i))
        return (i - self._start) // self._step

    def __len__(self):
        return len(six.moves.xrange(self._start, self._stop, self._step))

    def __reduce__(self):
        return (self.__class__, (self.start, self.stop, self.step))

    def __iter__(self):
        return range_iterator(self)

    def __repr__(self):
        return (__name__.split('.')[0] + '.' + self.__class__.__name__ +
                (str((self.start, self.stop)) if self.step == 1 else
                 str((self.start, self.stop, self.step))))



from .base import BaseItertool
from .iter_dispatch import iter_


class _grouper(BaseItertool):
    def __init__(self, value, iterator, groupby_obj):
        self._value = value
        self._groupby = groupby_obj
        self._key = self._groupby.key(self._value)
        self._initialized = False
        self._iterator = iterator
        self.stream_ended = False
        self._done = False

    def __next__(self):
        if not self._initialized:
            self._initialized = True
            return self._value
        else:
            if self._done:
                raise StopIteration
            try:
                value = next(self._iterator)
            except StopIteration:
                self.stream_ended = True
                self._done = True
                raise
            if self._groupby.key(value) != self._key:
                self.terminal_value = value
                self._done = True
                raise StopIteration
            return value


class groupby(BaseItertool):
    """groupby(iterable[, keyfunc]) -> create an iterator which returns
    (key, sub-iterator) grouped by each value of key(value).
    """
    def __init__(self, iterable, key=None):
        self._keyfunc = key
        self._iterator = iter_(iterable)
        self._current_key = self._initial_key = object()

    def key(self, value):
        if self._keyfunc is None:
            return value
        else:
            return self._keyfunc(value)

    def __next__(self):
        if not hasattr(self, '_current_grouper'):
            value = next(self._iterator)
            self._current_grouper = _grouper(value, self._iterator, self)
            return self.key(value), self._current_grouper
        else:
            while True:
                try:
                    next(self._current_grouper)
                except StopIteration:
                    break
            if self._current_grouper.stream_ended:
                raise StopIteration
            value = self._current_grouper.terminal_value
            self._current_grouper = _grouper(value, self._iterator, self)
            return self.key(value), self._current_grouper



"""Related things that aren't part of the standard library's `itertools`.

Currently home to picklable reimplementations of a few of the generators
from Matthew Rocklin's Toolz package.

Docstrings for `partition`, `partition_all`, and 'interleave' are lifted
wholesale from the Toolz documentation
<http://toolz.readthedocs.org/en/latest/>.
"""
import six
from .base import BaseItertool
from .map_zip import imap, izip_longest
from .iter_dispatch import iter_


class partition(BaseItertool):
    """Partition sequence into tuples of length n

    >>> list(partition(2, [1, 2, 3, 4]))
    [(1, 2), (3, 4)]

    If length of `seq` is not evenly divisible by `n`, the final
    tuple is dropped if `pad` is not specified, or filled to length
    `n` by `pad`:

    >>> list(partition(2, [1, 2, 3, 4, 5]))
    [(1, 2), (3, 4)]

    >>> list(partition(2, [1, 2, 3, 4, 5], pad=None))
    [(1, 2), (3, 4), (5, None)]

    See Also:
        partition_all
    """
    _NO_PAD = '__no_pad__'

    def __init__(self, n, seq, pad=_NO_PAD):
        self._n = n
        self._partition_all = partition_all(n, seq)
        self._pad = pad

    def __next__(self):
        items = next(self._partition_all)
        if len(items) < self._n:
            if self._pad != self._NO_PAD:
                items += (self._pad,) * (self._n - len(items))
            else:
                raise StopIteration
        return items


class partition_all(BaseItertool):
    """Partition all elements of sequence into tuples of length at most n

    The final tuple may be shorter to accommodate extra elements.

    >>> list(partition_all(2, [1, 2, 3, 4]))
    [(1, 2), (3, 4)]
    >>> list(partition_all(2, [1, 2, 3, 4, 5]))
    [(1, 2), (3, 4), (5,)]

    See Also:
        partition
    """
    def __init__(self, n, seq):
        self._n = n
        self._seq = iter_(seq)

    def __next__(self):
        items = []
        try:
            for _ in six.moves.xrange(self._n):
                items.append(next(self._seq))
        except StopIteration:
            pass
        if len(items) == 0:
            raise StopIteration
        return tuple(items)


class NoMoreItems(object):
    """Sentinel value for `equizip`. Do not use for any other purpose."""
    pass


class IterableLengthMismatch(ValueError):
    """Raised if an iterator passed to `equizip` is shorter than others."""
    pass


class equizip(izip_longest):
    """Like `izip_longest` but ensures the sequences are the same length.

    Raises :class:`IterableLengthMismatch` if one of the iterators
    terminates prematurely.
    """
    def __init__(self, *args):
        super(equizip, self).__init__(*args, fillvalue=NoMoreItems)

    def __next__(self):
        next_item = super(equizip, self).__next__()
        if any(value is NoMoreItems for value in next_item):
            raise IterableLengthMismatch
        return next_item


class interleave(BaseItertool):
    """Interleave a sequence of sequences

    >>> list(interleave([[1, 2], [3, 4]]))
    [1, 3, 2, 4]

    >>> ''.join(interleave(('ABC', 'XY')))
    'AXBYC'

    Both the individual sequences and the sequence of sequences may be infinite

    Returns a lazy iterator
    """
    def __init__(self, iterables, pass_exceptions=()):
        self._iters = imap(iter, iterables)
        self._more = []
        self._pass_exceptions = pass_exceptions

    def __next__(self):
        try:
            it = next(self._iters)
        except StopIteration:
            if len(self._more) == 0:
                raise
            else:
                self._iters = imap(iter, self._more)
                self._more = []
                return next(self)
        try:
            result = next(it)
            self._more.append(it)
            return result
        except (StopIteration,) + tuple(self._pass_exceptions):
            return next(self)


def roundrobin(*iterables):
    """Grab items from a collection of iterators in a round robin
    fashion until all are exhausted.

    >>> list(roundrobin('ABC', 'DEF', 'GH'))
    ['A', 'D', 'G', 'B', 'E', 'H', 'C', 'F']
    >>> list(roundrobin(xrange(2), xrange(5, 10), xrange(10, 12)))
    [0, 5, 10, 1, 6, 11, 7, 8, 9]

    """
    return interleave(iterables)
